@@ -2,7 +2,6 @@ use std::collections::HashMap;
 
 use bellframe::{AnnotBlock, Mask, Row, RowBuf};
 use index_vec::IndexVec;
-use itertools::Itertools;
 
 use super::{
     utils::{SNAP_FINISH_LABEL, SNAP_START_LABEL},
@@ -15,34 +14,27 @@ const LEADWISE_PREFIX: &str = "#";
 
 /// Creates a `Layout` where every course is exactly one lead long.
 pub fn leadwise(
-    methods: &[(bellframe::Method, String)],
-    calls: &[super::Call],
+    methods: &[super::Method],
     start_indices: Option<&[usize]>,
     end_indices: Option<&[usize]>,
 ) -> Result<Layout> {
-    super::utils::check_duplicate_shorthand(methods)?;
+    super::utils::check_duplicate_shorthands(methods)?;
 
     let stage = methods
         .iter()
-        .map(|(m, _shorthand)| m.stage())
+        .map(|m| m.method.stage())
         .max()
         .expect("Can't compute stage of 0 methods");
     let blocks = methods
         .iter()
-        .map(|(method, shorthand)| {
-            method
+        .map(|m| {
+            m.method
                 .first_lead()
-                .clone_map_annots_with_index(|i, _| (i == 0).then(|| shorthand.clone()))
+                .clone_map_annots_with_index(|i, _| (i == 0).then(|| m.shorthand.clone()))
         })
         .collect::<BlockVec<_>>();
 
-    // Compute the course head mask, which should only consist of fixed bells (e.g. `1xxxxxxx`
-    // for Plain Bob lead-head Major methods).
-    let calls_per_method = methods
-        .iter()
-        .map(|_| calls.iter().collect_vec())
-        .collect_vec();
-    let fixed_bells = super::utils::fixed_bells(methods, &calls_per_method, stage);
+    let fixed_bells = super::utils::fixed_bells(methods, stage);
     let lead_head_mask = Mask::fix_bells(stage, fixed_bells);
 
     let blks = blocks.as_raw_slice();
@@ -55,7 +47,7 @@ pub fn leadwise(
             blks,
         ),
         ends: start_or_ends(end_indices, "", SNAP_FINISH_LABEL, &lead_head_mask, blks),
-        links: links(methods, calls, &lead_head_mask),
+        links: links(methods, &lead_head_mask),
         blocks,
         stage,
     })
@@ -106,24 +98,21 @@ fn start_or_ends<I: index_vec::Idx>(
     locs
 }
 
-fn links(
-    methods: &[(bellframe::Method, String)],
-    calls: &[super::Call],
-    lead_head_mask: &Mask,
-) -> LinkVec<Link> {
+fn links(methods: &[super::Method], lead_head_mask: &Mask) -> LinkVec<Link> {
+    let mut call_starts: Vec<HashMap<&str, Vec<CallStart>>> = Vec::new();
     // Maps each lead label to where calls of that label can **end**
-    let mut call_starts: HashMap<&str, Vec<CallStart>> = HashMap::new();
     let mut call_ends: HashMap<&str, Vec<CallEnd>> = HashMap::new();
-    for (block_idx, (method, _)) in methods.iter().enumerate() {
-        let lead = method.first_lead();
+    for (block_idx, m) in methods.iter().enumerate() {
+        let lead = m.method.first_lead();
+        let mut call_starts_for_this_method = HashMap::new();
         for (row_idx_after, annot_row_after) in lead.annot_rows().enumerate() {
             if let Some(label) = annot_row_after.annot() {
                 let row_idx_before = (row_idx_after + lead.len() - 1) % lead.len();
                 let row_before = lead.get_row(row_idx_before).unwrap();
                 let row_after_plain = lead.get_row(row_idx_before + 1).unwrap();
 
-                call_starts
-                    .entry(label)
+                call_starts_for_this_method
+                    .entry(label.as_str())
                     .or_insert_with(Vec::new)
                     .push(CallStart {
                         row_idx: RowIdx::new(block_idx.into(), row_idx_before),
@@ -139,38 +128,41 @@ fn links(
                     });
             }
         }
+        call_starts.push(call_starts_for_this_method);
     }
 
     // Place calls between every `call_start` and every `call_end` of that lead label
     let mut links = Vec::new();
-    for call in calls {
-        let label = call.lead_location.as_str();
-        let starts = &call_starts[label];
-        let ends = &call_ends[label];
+    for (method_idx, method) in methods.iter().enumerate() {
+        for call in &method.calls {
+            let label = call.lead_location.as_str();
+            let starts = &call_starts[method_idx][label];
+            let ends = &call_ends[label];
 
-        for start in starts {
-            let mut row_after_call = start.row_before.clone();
-            call.place_not.permute(&mut row_after_call).unwrap();
-            for end in ends {
-                // Call
-                links.push(Link {
-                    from: start.row_idx,
-                    to: end.row_idx,
+            for start in starts {
+                let mut row_after_call = start.row_before.clone();
+                call.place_not.permute(&mut row_after_call).unwrap();
+                for end in ends {
+                    // Call
+                    links.push(Link {
+                        from: start.row_idx,
+                        to: end.row_idx,
 
-                    ch_mask: lead_head_mask.clone(),
-                    ch_transposition: &row_after_call * &end.inv_row,
+                        ch_mask: lead_head_mask.clone(),
+                        ch_transposition: &row_after_call * &end.inv_row,
 
-                    debug_name: call.debug_symbol.clone(),
-                    display_name: format!("[{}]", call.debug_symbol),
-                    weight: call.weight,
-                });
-                // Plain
-                links.push(plain_link(
-                    start.row_idx,
-                    end.row_idx,
-                    lead_head_mask,
-                    &start.row_after_plain * &end.inv_row,
-                ));
+                        debug_name: call.debug_symbol.clone(),
+                        display_name: format!("[{}]", call.debug_symbol),
+                        weight: call.weight,
+                    });
+                    // Plain
+                    links.push(plain_link(
+                        start.row_idx,
+                        end.row_idx,
+                        lead_head_mask,
+                        &start.row_after_plain * &end.inv_row,
+                    ));
+                }
             }
         }
     }
@@ -178,20 +170,20 @@ fn links(
     // Always add plain links at the end of every lead.  In nearly all cases, these will already
     // exist and be deduplicated, but if there are no calls at the end of each lead (e.g. in
     // link-cyclic or Stedman) then these will have to be generated separately.
-    for (method_idx_from, (method_from, _)) in methods.iter().enumerate() {
+    for (method_idx_from, method_from) in methods.iter().enumerate() {
         for (method_idx_to, _) in methods.iter().enumerate() {
             links.push(plain_link(
                 RowIdx {
                     block: method_idx_from.into(),
                     // - 1 to refer to the lead **end** not the lead **head**
-                    row: method_from.lead_len() - 1,
+                    row: method_from.method.lead_len() - 1,
                 },
                 RowIdx {
                     block: method_idx_to.into(),
                     row: 0,
                 },
                 lead_head_mask,
-                method_from.lead_head().to_owned(),
+                method_from.method.lead_head().to_owned(),
             ));
         }
     }
